import 'dart:developer';

import 'package:http/http.dart' as http;
import 'dart:convert' as convert;

late http.Response _httpResponse;

class HttpRequestException implements Exception {
  String message = '';
  String code = '';
  HttpRequestException({required this.code, required this.message});

  @override
  String toString() {
    String output = '[HTTP_REQUESTS - $code ] $message';
    return output;
  }
}

Future _responseConvert(val) async {
  _httpResponse = val;
  Response kponse = Response();
  return kponse;
}

String _headerCookies(Map<String, String> cookies) =>
    cookies.entries.map((e) => '${e.key}=${e.value}').join('; ');

class HttpRequests {
  /// An Method To Make Get HTTP Request
  /// Pass Url First Which Is Mandatory [ url like https://www.secanonm.in ] And You Can Pass Headers Using dictionary or map
  static get(String url,
      {Map<String, String>? headers, int timeout = 10}) async {
    http.Response response;
    var parseUrl = Uri();
    try {
      if (!url.startsWith('http')) {
        url = 'https://$url';
      }
      parseUrl = Uri.parse(url);
    } on FormatException {
      throw HttpRequestException(
        code: 'invalid url',
        message:
            'The requested url is invalid ( Please enter valid url or hostname ).',
      );
    }

    try {
      response = await http
          .get(parseUrl, headers: headers)
          .timeout(Duration(seconds: timeout), onTimeout: () {
        throw HttpRequestException(
          code: 'timeout',
          message:
              'A connection timeout ( The server at $url is taking too long to response. ) The requested site did not response to a connection request and http request has stopped waiting for a reply',
        );
      });
    } on FormatException {
      throw HttpRequestException(
        code: 'invalid url',
        message:
            'The requested url is invalid ( Please enter valid url or hostname ).',
      );
    } catch (e) {
      throw HttpRequestException(
        code: 'unreachable',
        message:
            'A network error (such as interrupted connection or host $url unreachable ) has occurred.',
      );
    }
    return _responseConvert(response);
  }

  /// An Method To Make Post HTTP Request
  /// Pass Url First Which Is Mandatory [ url like https://www.secanonm.in ] And You Can Pass header And Data Using dictionary or map
  static post(String url,
      {Map<String, String>? headers,
      Map<String, dynamic>? data,
      Map<String, String>? cookies,
      int timeout = 10}) async {
    http.Response response;

    var parseUrl = Uri();
    try {
      if (!url.startsWith('http')) {
        url = 'https://$url';
      }
      parseUrl = Uri.parse(url);
    } on FormatException {
      throw HttpRequestException(
        code: 'invalid url',
        message:
            'The requested url is invalid ( Please enter valid url or hostname ).',
      );
    }

    if (cookies != null) {
      if (data != null) {
        data['Cookie'] = _headerCookies(cookies);
      } else {
        data = {'Cookie': _headerCookies(cookies)};
      }
    }

    try {
      response = await http
          .post(parseUrl, body: data, headers: headers)
          .timeout(Duration(seconds: timeout), onTimeout: () {
        throw HttpRequestException(
          code: 'timeout',
          message:
              'A connection timeout ( The server at $url is taking too long to response. ) The requested site did not response to a connection request and http request has stopped waiting for a reply',
        );
      });
    } on FormatException {
      throw HttpRequestException(
        code: 'invalid url',
        message:
            'The requested url is invalid ( Please enter valid url or hostname ).',
      );
    } catch (e) {
      throw HttpRequestException(
        code: 'unreachable',
        message:
            'A network error (such as interrupted connection or host $url unreachable ) has occurred.',
      );
    }

    return await _responseConvert(response);
  }
}

Map<String, String> _httpStatusCodeMap = {
  // 2XX Success
  '200': 'Success',
  '201': 'Created',
  '204': 'No Content',

  // 3XX Redirection
  '301': 'Move Permanently',
  '302': 'Found',
  '304': 'Not Modified',
  '307': 'Temporary Redirect',
  '308': 'Permanent Redirect',

  // 4XX Client Error
  '400': 'Bad Request',
  '401': 'Unauthorized',
  '403': 'Forbidden',
  '404': 'Not Found',
  '405': 'Method Not Allowed',
  '406': 'Not Accepted',
  '408': 'Request Timeout',
  '409': 'Conflict',
  '451': 'Unavailable For Legal Reasons',

  // 5XX Client Error
  '500': 'Internal Server Error',
  '501': 'Not Implemented',
  '502': 'Bad Gateway',
  '503': 'Service Unavailable',
  '504': 'Gateway Timeout',
  '521': 'Web Server Is Down',
  '522': 'Connection Timed Out',
  '523': 'Origin Is Unreachable',
};

/// Simply Class To Handle Response
class Response {
  /// Simply Return Response code
  bool get hasError =>
      (400 <= _httpResponse.statusCode) && (_httpResponse.statusCode >= 100);

  /// It return success status of request ,if request is complete than it return true else return false
  bool get success => !hasError;

  /// It return HTTP status Code with description like `200 - success`
  get status =>
      "${_httpResponse.statusCode} - ${_httpStatusCodeMap[_httpResponse.statusCode.toString()] ?? 'unknown'}";

  /// It return HTTP status Code of response like 200 , 301, 502
  get statusCode => _httpResponse.statusCode;

  /// It return Content Type of response like json or text
  String get contentType => _httpResponse.headers['content-type'] ?? "undefine";

  /// It return content of response
  String get content => convert.utf8.decode(bytes, allowMalformed: true);

  /// It return json response in format of Map { like in python dict }
  Map<String, dynamic> get json {
    Map<String, dynamic> _helo;
    try {
      _helo = convert.jsonDecode(_httpResponse.body) as Map<String, dynamic>;
    } catch (e) {
      log('Failed to convert JSON', name: 'HTTP_REQUESTS');
      _helo = {"status": "failed"};
    }
    return _helo;
  }

  /// It return response header in format of Map { like in python dict }
  get headers => _httpResponse.headers;

  /// It return content of response simply
  get response => _httpResponse.body;

  /// It return content length of response simply
  get contentLength => _httpResponse.contentLength;

  /// It return Redirection Status of response simply
  get isRedirect => _httpResponse.isRedirect;

  /// It return bytes comprising the body of response
  List<int> get bytes => _httpResponse.bodyBytes;
}
